package net.i2p.router.crypto.ratchet;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.*;

import net.i2p.I2PAppContext;
import net.i2p.crypto.EncType;
import net.i2p.crypto.KeyFactory;
import net.i2p.crypto.KeyPair;
import net.i2p.data.PrivateKey;
import net.i2p.data.PublicKey;
import net.i2p.router.RouterContext;
import net.i2p.util.I2PThread;
import net.i2p.util.Log;
import net.i2p.util.SystemVersion;
import net.i2p.router.JobImpl;
import net.i2p.router.Job;

/**
 *  Elligator2 for X25519 keys.
 *
 *  Try to keep DH pairs at the ready.
 *  It's important to do this in a separate thread, because if we run out,
 *  the pairs are generated in the NTCP Pumper thread,
 *  and it can fall behind.
 *
 *  @since 0.9.44 from X25519KeyFactory
 */
public class Elg2KeyFactory implements KeyFactory {

    private final RouterContext _context;
    private final Log _log;
    private final int _minSize;
    private final int _maxSize;
    private final int _calcDelay;
    private final LinkedBlockingQueue<Elg2KeyPair> _keys;
    private volatile boolean _isRunning;
    private AtomicBoolean _jobRunning = new AtomicBoolean();
    private AtomicReference<Job> _job = new AtomicReference<Job>(null);

    private final static String PROP_DH_PRECALC_MIN = "crypto.edh.precalc.min";
    private final static String PROP_DH_PRECALC_MAX = "crypto.edh.precalc.max";
    private final static String PROP_DH_PRECALC_DELAY = "crypto.edh.precalc.delay";
    private final static int DEFAULT_DH_PRECALC_MIN = 10;
    private final static int DEFAULT_DH_PRECALC_MAX = 30;
    private final static int DEFAULT_DH_PRECALC_DELAY = 25;

    public  Elg2KeyFactory(RouterContext ctx) {
        _context = ctx;
        _log = ctx.logManager().getLog(Elg2KeyFactory.class);
        ctx.statManager().createRateStat("crypto.EDHGenerateTime", "How long it takes to create x and X", "Encryption", new long[] { 60*60*1000 });
        ctx.statManager().createRateStat("crypto.EDHUsed", "Need a DH from the queue", "Encryption", new long[] { 60*60*1000 });
        ctx.statManager().createRateStat("crypto.EDHReused", "Unused DH requeued", "Encryption", new long[] { 60*60*1000 });
        ctx.statManager().createRateStat("crypto.EDHEmpty", "DH queue empty", "Encryption", new long[] { 60*60*1000 });

        // add to the defaults for every 128MB of RAM, up to 512MB
        long maxMemory = SystemVersion.getMaxMemory();
        int factor = (int) Math.max(1l, Math.min(4l, 1 + (maxMemory / (128*1024*1024l))));
        int defaultMin = DEFAULT_DH_PRECALC_MIN * factor;
        int defaultMax = DEFAULT_DH_PRECALC_MAX * factor;
        _minSize = ctx.getProperty(PROP_DH_PRECALC_MIN, defaultMin);
        _maxSize = ctx.getProperty(PROP_DH_PRECALC_MAX, defaultMax);
        _calcDelay = ctx.getProperty(PROP_DH_PRECALC_DELAY, DEFAULT_DH_PRECALC_DELAY);

        if (_log.shouldLog(Log.DEBUG))
            _log.debug("EDH Precalc (minimum: " + _minSize + " max: " + _maxSize + ", delay: "
                       + _calcDelay + ")");
        _keys = new LinkedBlockingQueue<Elg2KeyPair>(_maxSize);
    }
        
    /**
     *  Note that this stops the singleton precalc thread.
     *  You don't want to do this if there are multiple routers in the JVM.
     *  Fix this if you care. See Router.shutdown().
     */
    public void shutdown() {
        _isRunning = false;
        _keys.clear();
    }

    public void start() {
        _isRunning = true;
    }

    final class EDHPrecalcJob extends JobImpl {

        private EDHPrecalcJob() {
            super(_context);
        }

        public void runJob() {
            if (!_isRunning) return;
            while (_maxSize > _keys.size())
                addKeys(precalc());
            _jobRunning.set(false);
        }

        public String getName() { return "EDH Session Key Precalc"; }
    }

    /**
     * Pulls a prebuilt keypair from the queue,
     * or if not available, construct a new one.
     */
    public Elg2KeyPair getKeys() {
        _context.statManager().addRateData("crypto.EDHUsed", 1);
        Elg2KeyPair rv = _keys.poll();
        if (rv == null) {
            _context.statManager().addRateData("crypto.EDHEmpty", 1);
            rv = precalc();
            if (_job.get() == null)
                _job.compareAndSet(null, new EDHPrecalcJob());
        }
        if (_jobRunning.compareAndSet(false, true))
            _context.jobQueue().addJob(_job.get());
        return rv;
    }

    private Elg2KeyPair precalc() {
        long start = _context.clock().now();
        KeyPair rv;
        byte[] enc;
        int i = 0;
        do {
            rv = _context.keyGenerator().generatePKIKeys(EncType.ECIES_X25519);
            enc = Elligator2.encode(rv.getPublic(), _context.random().nextBoolean());
            i++;
        } while (enc == null);
        long diff = _context.clock().now() - start;
        _context.statManager().addRateData("crypto.EDHGenerateTime", diff);
        if (_log.shouldLog(Log.DEBUG))
            _log.debug("Took " + i + " tries and " + diff + "ms to generate local DH value");
        return new Elg2KeyPair(rv.getPublic(), rv.getPrivate(), enc);
    }

    /**
     * Return an unused DH key builder
     * to be put back onto the queue for reuse.
     */
    public void returnUnused(Elg2KeyPair kp) {
/*
        _context.statManager().addRateData("crypto.EDHReused", 1);
        _keys.offer(kp);
*/
    }

    /** @return true if successful, false if full */
    private final boolean addKeys(Elg2KeyPair kp) {
        return _keys.offer(kp);
    }

    private final int getSize() {
        return _keys.size();
    }

}
